//----------------------------------------------------------------------------

//   Wake   RS-485.

//----------------------------------------------------------------------------

#include <Main.h>
#include <Port.h>
#include <Wake.h>

//----------------------------- : -----------------------------------

#define BAUD 9600        // 
#define RS485            //   RS485
#define TX_DELAY 20      // , 
#define ADDR_KEY 0xBEDA  //   

#define UBRRV (int)((FCLK * 1000000.0)/(16.0 * BAUD) - 0.5) //

//------------------------------- : --------------------------------

char Rx_Sta,        //   
     Rx_Pre,        //  
     Rx_Add,        //,    
     Rx_Cmd,        // 
     Rx_Nbt,        //    
     Rx_Dat[FRAME], //  
     Rx_Crc,        //   
     Rx_Ptr;        //    

char Command;          //   

char Tx_Sta,        //   
     Tx_Pre,        //  
     Tx_Add,        //,   
     Tx_Cmd,        //,   
     Tx_Nbt,        //    
     Tx_Dat[FRAME], //   
     Tx_Crc,        //   
     Tx_Ptr;        //    

static char TxTimer; //     RS-485

__no_init __eeprom int EAddr; //  EEPROM

//-------------------------------- : ----------------------------------

#pragma vector = USART_TXC_vect
__interrupt void PC_Tx(void);    //   
#pragma vector = USART_RXC_vect
__interrupt void PC_Rx(void);    //   
void Do_Crc8(char b, char *crc); //  
__monitor void SaveAddress(char addr); //   EEPROM
__monitor char ReadAddress(void); //   EEPROM

//------------------------  UART: -------------------------------

void Port_Init(void)
{
  UCSRC = (1<<UCSZ1) | (1<<UCSZ0); //8- 
  UBRRL = LO(UBRRV); //  
  UBRRH = HI(UBRRV);
  //     ,    
  UCSRB = (1<<RXCIE) | (1<<TXCIE) | (1<<RXEN) | (1<<TXEN);

  Rx_Add = ReadAddress(); //   EEPROM
  Tx_Add = Rx_Add;        //      
  Rx_Sta = WAIT_FEND;     // 
  Tx_Sta = SEND_IDLE;     //   
  Command = CMD_NOP;      //   
}

//-----------------------    EEPROM: ------------------------

void SaveAddress(char addr)
{
  EAddr = addr;
}

//-------------------------    EEPROM: -------------------------

char ReadAddress(void)
{
  char a = EAddr;
  if(a > 127) a = 0;
  return(a);
}

//-----------------------   : -----------------------

bool Port_Set_Address(unsigned int sig, char addr)
{
  if(sig == ADDR_KEY && addr < 128)
  {
    Rx_Add = addr;
    Tx_Add = addr;
    SaveAddress(addr);
    return(1);
  }
  return(0);
}

//------------------------   : -------------------------

char Port_Get_Address(void)
{
  return(Rx_Add);
}

//-------------------  UART   : --------------------

#pragma vector = USART_RXC_vect
__interrupt void PC_Rx(void)
{
  char error_flags = (UCSRA & (1<<FE)); //  
  char data_byte = UDR;               // 

  if(error_flags)                     //     
  {
    Rx_Sta = WAIT_FEND;               //  
    Command = CMD_ERR;                //  
    return;
  }

  if(data_byte == FEND)               //   ,
  {
    Rx_Pre = data_byte;               //  -,
    Rx_Crc = CRC_INIT;                // CRC,
    Rx_Sta = WAIT_ADDR;               //  ,
    Do_Crc8(data_byte, &Rx_Crc);      // CRC,
    return;                           //
  }

  if(Rx_Sta == WAIT_FEND)             //----->   FEND,
    return;                           // 

  char Pre = Rx_Pre;                  //  -
  Rx_Pre = data_byte;                 // -
  if(Pre == FESC)                     // -  FESC,
  {
    if(data_byte == TFESC)            //    TFESC,
      data_byte = FESC;               //    FESC
    else if(data_byte == TFEND)       //    TFEND,
           data_byte = FEND;          //    FEND
         else
         {
           Rx_Sta = WAIT_FEND;        //     ,
           Command = CMD_ERR;         //  FESC, 
           return;
         }
  }
  else
  {
    if(data_byte == FESC)             //    FESC,  
      return;                         //  -
  }

  switch(Rx_Sta)
  {
  case WAIT_ADDR:                     //----->   
    {
      if(data_byte & 0x80)            //  7    ,   
      {
        data_byte = data_byte & 0x7F; //  7,   
        if(data_byte == 0 || data_byte == Rx_Add) //    ,
        {
          Do_Crc8(data_byte, &Rx_Crc); //  CRC 
          Rx_Sta = WAIT_CMD;          //   
          break;
        }
        Rx_Sta = WAIT_FEND;           //  ,   
        break;
      }                               //  7   , 
      Rx_Sta = WAIT_CMD;              //    
    }
  case WAIT_CMD:                      //----->   
    {
      if(data_byte & 0x80)            //  7 
      {
        Rx_Sta = WAIT_FEND;           //  7   ,
        Command = CMD_ERR;            // 
        break;
      }
      Rx_Cmd = data_byte;             // 
      Do_Crc8(data_byte, &Rx_Crc);    // CRC
      Rx_Sta = WAIT_NBT;              //    
      break;
    }
  case WAIT_NBT:                      //----->    
    {
      if(data_byte > FRAME)           //   > FRAME,
      {
        Rx_Sta = WAIT_FEND;
        Command = CMD_ERR;            // 
        break;
      }
      Rx_Nbt = data_byte;
      Do_Crc8(data_byte, &Rx_Crc);    // CRC
      Rx_Ptr = 0;                     //  
      Rx_Sta = WAIT_DATA;             //   
      break;
    }
  case WAIT_DATA:                     //----->   
    {
      if(Rx_Ptr < Rx_Nbt)             //    ,
      {
        Rx_Dat[Rx_Ptr++] = data_byte; //   ,
        Do_Crc8(data_byte, &Rx_Crc);  // CRC
        break;
      }
      if(data_byte != Rx_Crc)         //   ,   CRC
      {
        Rx_Sta = WAIT_FEND;           // CRC  ,
        Command = CMD_ERR;            // 
        break;
      }
      Rx_Sta = WAIT_FEND;             //  ,
      Command = Rx_Cmd;               //   
      break;
    }
  }
}

//------------------  UART   : -------------------

#pragma vector = USART_TXC_vect
__interrupt void Tx_PC(void)
{
  if(Tx_Sta == SEND_FEND) return;     //  

  char data_byte;
  if(Tx_Pre == FEND)                  //  ,
  {
    data_byte = TFEND;                // TFEND  FEND
    Tx_Pre = data_byte;
    UDR = data_byte;
    return;
  }
  if(Tx_Pre == FESC)                  //  ,
  {
    data_byte = TFESC;                // TFESC  FESC
    Tx_Pre = data_byte;
    UDR = data_byte;
    return;
  }

  switch(Tx_Sta)
  {
  case SEND_ADDR:                     //----->  
    {
      if(Tx_Add)                      //    ,
      {
        data_byte = Tx_Add;
        Do_Crc8(data_byte, &Tx_Crc);  //  CRC,
        data_byte |= 0x80;            // 7    
        Tx_Pre = data_byte;           // -
        if(data_byte == FEND || data_byte == FESC)
          data_byte = FESC;           // FESC,   
        UDR = data_byte;
        Tx_Sta = SEND_CMD;
        return;
      }
      else
        Tx_Sta = SEND_CMD;            //   
    }
  case SEND_CMD:                      //----->  
    {
      data_byte = Tx_Cmd & 0x7F;
      Tx_Sta = SEND_NBT;
      break;
    }
  case SEND_NBT:                      //----->   
    {
      data_byte = Tx_Nbt;
      Tx_Sta = SEND_DATA;
      Tx_Ptr = 0;                     //    
      break;
    }
  case SEND_DATA:                     //----->  
    {
      if(Tx_Ptr < Tx_Nbt)
        data_byte = Tx_Dat[Tx_Ptr++];
      else
      {
        data_byte = Tx_Crc;           // CRC
        Tx_Sta = SEND_CRC;
      }
      break;
    }
  default:
    {
#ifdef RS485
      Port_DIR_0;                     // RS-485  
#endif
      Tx_Sta = SEND_IDLE;             //  
      return;
    }
  }

  Do_Crc8(data_byte, &Tx_Crc);        // CRC
  Tx_Pre = data_byte;                 // -
  if(data_byte == FEND || data_byte == FESC)
    data_byte = FESC;                 // FESC,   
  UDR = data_byte;
}

//---------------------   : ------------------------

void Do_Crc8(char b, char *crc)
{
  for(char i = 0; i < 8; b = b >> 1, i++)
    if((b ^ *crc) & 1) *crc = ((*crc ^ 0x18) >> 1) | 0x80;
     else *crc = (*crc >> 1) & ~0x80;
}

//------------------------   : ---------------------------

void Port_Tx_Frame(void)
{
#ifdef RS485
  TxTimer = ms2sys(TX_DELAY);         //  RS-485,   
#else
  TxTimer = 0;                        //  RS-232,   
#endif
  Tx_Sta = SEND_FEND;                 //   
}

//------------------     : ---------------------

void Port_Tx_Exe(bool t)
{
  if(Tx_Sta == SEND_FEND)             //   
  {
    if(TxTimer == 0)                  //   
    {
      char data_byte = FEND;
      Tx_Crc = CRC_INIT;              // CRC,
      Do_Crc8(data_byte, &Tx_Crc);    // CRC
      Tx_Sta = SEND_ADDR;             //   
      Tx_Pre = TFEND;
#ifdef RS485
      Port_DIR_1;                     // RS-485  
#endif
      UDR = data_byte;                // FEND
    }
    else
      if(t) TxTimer--;                //  
  }
}

//----------------------------------------------------------------------------
